//***********************************************************************************//
//*	Project:																		*//
//*																					*//
//*	File:		`$INSTANCE_NAME`.c			 										*//
//*	Version:	1.0																	*//
//*																					*//
//*	Author:		Bob Marlowe															*//
//*	(C) 2018	Jörg Meier Software - Entwicklung briefe@jmeiersoftware.de			*//
//***********************************************************************************//
//***********************************************************************************//
//*																					*//
//*		Principles of Operation														*//
//*																					*//
//*		Driver for up to 10 servos for 1ms to 2ms pulse control						*//
//*																					*//
//*		A single PWM with interrupts for TC and CC (indicates pulse end) 			*//
//*		is used. The servo outputs are changed by the software, so there is 		*//
//*		only one pulse active at the time.											*//
//*		Pulses are repeated every 20ms, so after all servos are handled a delay		*//
//*		is appended to maintain the 20 ms (50Hz) interval							*//
//*																					*//
//***********************************************************************************//

//***********************************************************************************//
//**************** Libraries ********************************************************//
//***********************************************************************************//
#include <project.h>
#include <`$INSTANCE_NAME`.h>

//***********************************************************************************//
//**************** Definitions and Macros *******************************************//
//***********************************************************************************//
#define PWMClock	24000			//	Clock for PWM in kHz
#define One_ms		24000			//	Clock counts for 1 ms

//***********************************************************************************//
//******************* Structs, Typedefs and Enums ***********************************//
//***********************************************************************************//
typedef void (*PinWrite)(uint8);			//	Pointer to the pin write function


//***********************************************************************************//
//******************* Data Area *****************************************************//
//***********************************************************************************//
static uint16 ServoValues[`$NUM_SERVOS`];						//	Array for the PWM compare values

PinWrite const WritePins[] = 
					{
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_0_H
						&`$INSTANCE_NAME`_Servo_0_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_1_H
						&`$INSTANCE_NAME`_Servo_1_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_2_H
						&`$INSTANCE_NAME`_Servo_2_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_3_H
						&`$INSTANCE_NAME`_Servo_3_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_4_H
						&`$INSTANCE_NAME`_Servo_4_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_5_H
						&`$INSTANCE_NAME`_Servo_5_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_6_H
						&`$INSTANCE_NAME`_Servo_6_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_7_H
						&`$INSTANCE_NAME`_Servo_7_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_8_H
						&`$INSTANCE_NAME`_Servo_8_Write,
#endif
#ifdef CY_PINS_`$INSTANCE_NAME`_Servo_9_H
						&`$INSTANCE_NAME`_Servo_9_Write,
#endif
					};
volatile uint8 `$INSTANCE_NAME`_CycleDone;					//	Indicator for a completed cycle

//***********************************************************************************//
//****************** Prototypes *****************************************************//
//***********************************************************************************//
static CY_ISR_PROTO(`$INSTANCE_NAME`_PWM_Handler);			//	Interrupt handler

//***********************************************************************************//
//**************** Program Area *****************************************************//
//***********************************************************************************//
void `$INSTANCE_NAME`_Start(void)							//	Initialize all servos and set mid-position as default
{
uint8 ii;
	for(ii = 0; ii < `$INSTANCE_NAME`_NumServos; ii++)
	{
		`$INSTANCE_NAME`_SetServo(ii,50);								//	All servos into mid position
	}
	`$INSTANCE_NAME`_ServoPWM_Init();									//	Initialize the PWM
	`$INSTANCE_NAME`_ServoInt_StartEx(`$INSTANCE_NAME`_PWM_Handler);								//	Setup the interrupt handler
	`$INSTANCE_NAME`_CycleDone = 0;										//	Indicate no cycle completed
}
//***********************************************************************************//

void `$INSTANCE_NAME`_MoveServos(void)									//	Start moving of all servos
{
	`$INSTANCE_NAME`_ServoPWM_WriteCompare(ServoValues[0]);				//	Initial servo 0 pulse width
	WritePins[0](1);													//	Set servo 0 output to one 
	`$INSTANCE_NAME`_ServoPWM_Start();									//	Engines start
}
//***********************************************************************************//

void `$INSTANCE_NAME`_SetServo(uint8 ServoNumber,uint8 Percent)			//	Set servo position between 0 and 100%
{
uint8 IntState;
uint32 Value;
	if(ServoNumber >= `$INSTANCE_NAME`_NumServos) return;				//	Invalid servo number
	if(Percent > 100) Percent = 100;									//	Correction
	Value = One_ms + (One_ms / 100) * Percent;							//	Giving a range of 1ms .. 2ms#
	IntState = CyEnterCriticalSection();								//	Just iin case of a conflict
	ServoValues[ServoNumber] = Value;									//	Set the result
	CyExitCriticalSection(IntState);									//	Allow interrupts back again
}
//***********************************************************************************//

uint8 `$INSTANCE_NAME`_GetServo(uint8 Servonumber)						//	Get currently set percentage of servo
{
uint32 Value;
	if(Servonumber >= `$INSTANCE_NAME`_NumServos) return 0;				//	Invalid servo returns 0
	Value = ServoValues[Servonumber] - One_ms;							//	Conversion
	return (Value * 100) / One_ms;										//	
}
//***********************************************************************************//

CY_ISR(`$INSTANCE_NAME`_PWM_Handler)									//	Interrupt handler
{
static uint8 ServoCounter = 0;											//	Initial value 
static uint8 AppendCount = 0;											//	Counter for appending timeslices to get 20 ms or 50 Hz
uint32 IntCause;														//	Interrupt cause can be tc or compare
uint16 WriteValue;
	IntCause = `$INSTANCE_NAME`_ServoPWM_GetInterruptSource();			//	Get interrupt bits
	if(`$INSTANCE_NAME`_ServoPWM_INTR_MASK_CC_MATCH & IntCause)			//	Pulse high phase expired?
	{
		if(ServoCounter < `$INSTANCE_NAME`_NumServos)					//	Avoid the 0xff value
			WritePins[ServoCounter](0);									//	Set actual servo pulse to 0, pulse finished
		`$INSTANCE_NAME`_ServoPWM_ClearInterrupt(`$INSTANCE_NAME`_ServoPWM_INTR_MASK_CC_MATCH);	//	Clear the compare interrupt
		return;
	}
	
//******** We get here only when we have got a TC interrupt	
	`$INSTANCE_NAME`_ServoPWM_ClearInterrupt(`$INSTANCE_NAME`_ServoPWM_INTR_MASK_TC);				//	Clear the tc interrupt
		
	if(AppendCount > 0)													//	We need to append 2ms time
	{
		AppendCount--;													//	We are appending some time to get a repeat rate of 20ms
		if(AppendCount == 0) `$INSTANCE_NAME`_CycleDone = 1;			//	Indicate last cycle
		return;
	}
	
//******** We get here only when last pulse was finished	
	ServoCounter++;														//	Next channel
	if(ServoCounter >= `$INSTANCE_NAME`_NumServos)						//	Appending pulses?
	{
		ServoCounter = 0xff;											//	Adjust to valid value in next cycle
		AppendCount = MaxServos - `$INSTANCE_NAME`_NumServos;			//	Number of 2 ms times to append. results in a 20ms repeatition rate
		WriteValue = 2 * One_ms;										//	2 ms duration
	}
	else
	{
		WriteValue = ServoValues[ServoCounter];							//	New value from current channel
		WritePins[ServoCounter](1);										//	Set actual servo output to one 
	}
	
	`$INSTANCE_NAME`_ServoPWM_WriteCompare(WriteValue);					//	New pulse width
}
//***********************************************************************************//
